class Todo:
    #<<description, status (done, not done), id number>>

def load_todos() -> list[Todo]:
    #<<load from ~/.todos.json>>

todos = load_todos()

def save_todos(todos: list[Todo]) -> None:
    #<<save todos>>

def add_todo(description):
    #<<add todo and save>>

def remove_todo(id):
    #<<remove todo and save>>

def print_todos():
    #<<print todos>>

#<<command line interface using argparse>>



import os
import json
import hashlib

def key_to_hash(key: dict) -> str:
    hasher = hashlib.sha256()

    hasher.update(json.dumps(key).encode("utf-8"))

    return hasher.hexdigest()

class CacheKeyMissingError(Exception):
    pass

class DiskCache:
    def __init__(self, cache_dir):
        os.makedirs(cache_dir, exist_ok=True)

        self.cache_dir = cache_dir

    def path(self, key: dict) -> str:
        return os.path.join(
            self.cache_dir,
            key_to_hash(key) + ".json",
        )

    def set(self, key: dict, value: object) -> None:
        with open(self.path(key), "w") as file:
            json.dump(value, file)

    def get(self, key: dict) -> object:
        try:
            with open(self.path(key), "r") as file:
                return json.load(file)
        except FileNotFoundError:
            raise CacheKeyMissingError()

default = DiskCache(
    cache_dir=os.environ.get(
        "MACCARONE_CACHE_DIR",
        os.path.expanduser("~/.cache/maccarone"),
    ),
)



import re
import logging

from dataclasses import dataclass

from maccarone.openai import complete_chat_with_cache

logger = logging.getLogger(__name__)

class Piece:
    pass

@dataclass
class PresentPiece(Piece):
    text: str

@dataclass
class MissingPiece(Piece):
    indent: str
    guidance: str

def raw_source_to_pieces(input: str) -> list[Piece]:
    missing = re.finditer(r"(\n(?P<il>[ \t]*))?(?P<lm>#<<)(?P<gd>.+?)(?P<rm>>>)", input, re.DOTALL)
    position = 0
    pieces = []

    for piece in missing:
        left = piece.start("lm")

        pieces += [
            PresentPiece(text=input[position:left]),
            MissingPiece(
                indent=piece.group("il") or "",
                guidance=piece.group("gd"),
            ),
        ]

        position = piece.end("rm")

    pieces += [PresentPiece(text=input[position:])]

    return pieces

def raw_pieces_to_tagged_input(raw_pieces: list[Piece]) -> str:
    tag_source = ""
    id = 0

    for piece in raw_pieces:
        match piece:
            case PresentPiece(text):
                tag_source += text
            case MissingPiece(indent, guidance):
                tag_source += f'# <write_this id="{id}">\n{indent}# {guidance}\n{indent}# </>'
                id += 1
            case _:
                raise TypeError("unknown piece type", piece)

    logger.debug("tagged input ↓\n%s", tag_source)

    return tag_source

def tagged_input_to_tagged_output(tagged_input: str) -> str:
    system_prompt = """
You are an expert programmer working on contract. Your client has written a partial program, but left pieces for you to complete. They have marked those with `<write_this>` tags inside Python comments, e.g.:

```
def add_two_numbers(x, y):
    # <write_this id="0">
    # add the two numbers
    # </>

# <write_this id="1">
# add two numbers from command line args, using argparse
# </>
```

You should produce a document that provides a `<completed>` tag for each missing piece, e.g.:

```
<completed id="0">
return x + y
</>
<completed id="1">
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("x", type=int)
parser.add_argument("y", type=int)
args = parser.parse_args()
return add_two_numbers(args.x, args.y)
</>
```

This formatting is very important. The client uses a custom tool to process your work product, and their tool requires this format. Follow this format exactly and do not copy anything outside a `<write_this>` tag.
"""
    chat_messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": tagged_input},
    ]
    tagged_output = complete_chat_with_cache(chat_messages)

    logger.debug("tagged output ↓\n%s", tagged_output)

    return tagged_output

def tagged_output_to_completed_pieces(tagged_output: str) -> dict[int, str]:
    pattern = re.compile(r'<completed id="(?P<id>\d+)">\n(?P<content>.+?)</>', re.DOTALL)
    matches = pattern.finditer(tagged_output)
    completed = {int(m.group("id")): m.group("content") for m in matches}

    return completed

def pieces_to_final_source(
        raw_pieces: list[Piece],
        completed_pieces: dict[int, str],
    ) -> str:
    id = 0
    final_source = ""

    for raw in raw_pieces:
        match raw:
            case PresentPiece(text):
                final_source += text
            case MissingPiece(indent):
                completed = completed_pieces[id]
                final_source += indent.join(completed.splitlines(True))
                id += 1
            case _:
                raise TypeError("unknown piece type", raw)

    logger.debug("final source ↓\n%s", final_source)

    return final_source

def preprocess_maccarone(raw_source: str) -> str:
    raw_pieces = raw_source_to_pieces(raw_source)
    tagged_input = raw_pieces_to_tagged_input(raw_pieces)
    tagged_output = tagged_input_to_tagged_output(tagged_input)
    completed_pieces = tagged_output_to_completed_pieces(tagged_output)
    final_source = pieces_to_final_source(raw_pieces, completed_pieces)

    return final_source



import os
import sys
import logging

from importlib.abc import (
    MetaPathFinder,
    SourceLoader,
)
from importlib.machinery import ModuleSpec
from fnmatch import fnmatchcase

from maccarone.preprocessor import preprocess_maccarone

class ImportFinder(MetaPathFinder):
    def __init__(
            self,
            py_string_matching: bool,
            include_pattern: str | None,
            exclude_pattern: str | None,
        ) -> None:
        self.py_string_matching = py_string_matching
        self.include_pattern = include_pattern
        self.exclude_pattern = exclude_pattern

    def find_spec(self, fullname, path, target=None):
        # check against module name patterns, if configured
        if self.exclude_pattern is not None:
            if fnmatchcase(fullname, self.exclude_pattern) is not None:
                return None

        if self.include_pattern is not None:
            if fnmatchcase(fullname, self.include_pattern) is None:
                return None

        # module name looks ok; check for maccarone snippets
        def make_modulespec(filename):
            return ModuleSpec(
                fullname,
                ImportLoader(fullname, filename),
                origin=filename,
                is_package=False
            )

        if path is None or path == '':
            path = [os.getcwd()]  # top level import 

        for entry in path:
            parts = fullname.split(".")
            basename = parts[-1]
            package_path = parts[1:-1]
            base_filename = os.path.join(entry, *package_path, basename)
            py_filename = base_filename + '.py'
            mn_filename = base_filename + '.mn.py'

            if os.path.exists(mn_filename):
                return make_modulespec(mn_filename)
            elif self.py_string_matching and os.path.exists(py_filename):
                with open(py_filename, "rt") as file:
                    if "#<<" in file.read():
                        return make_modulespec(py_filename)

        return None  # we don't know how to import this

class ImportLoader(SourceLoader):
    def __init__(self, fullname, path):
        self.fullname = fullname
        self.path = path

    def get_filename(self, fullname):
        return self.path

    def get_data(self, filename):
        with open(self.path, 'r') as file:
            in_source = file.read()

        return preprocess_maccarone(in_source)

def enable(
        py_string_matching=True,
        include_pattern=None,
        exclude_pattern=None,
    ):
    """Exclude patterns take precedence over include patterns."""

    import_finder = ImportFinder(
        py_string_matching=py_string_matching,
        include_pattern=include_pattern,
        exclude_pattern=exclude_pattern,
    )

    sys.meta_path.insert(0, import_finder)

    if os.environ.get("MACCARONE_LOGGING", False):
        logging.basicConfig(level=logging.DEBUG)



import os
import logging

from typing import (
    Callable,
    Iterable,
    cast,
)

import openai

from openai import ChatCompletion

from maccarone.caching import (
    default as cache,
    CacheKeyMissingError,
)

logger = logging.getLogger(__name__)

openai.api_key = os.getenv("OPENAI_API_KEY")

def complete_chat(
        messages: list[dict[str, str]],
        model="gpt-4",
        on_token: Callable[[int], None] = lambda p: None,
    ) -> str:
    responses = cast(
        Iterable[ChatCompletion],
        ChatCompletion.create(
            model=model,
            messages=messages,
            stream=True,
            temperature=0.0,
        ),
    )
    completion = ""

    logger.info("completing: %r", messages)

    for (i, partial) in enumerate(responses):
        delta = partial.choices[0].delta

        try:
            completion += str(delta.content)
        except AttributeError as error:
            pass

        on_token(i)

    logger.info("completion: %r", completion)

    return completion

def complete_chat_with_cache(
        messages: list[dict[str, str]],
        model="gpt-4",
    ) -> str:
    try:
        return cache.get(messages)
    except CacheKeyMissingError:
        completion = complete_chat(messages, model=model)

        cache.set(messages, completion)

        return completion



import pytest

from textwrap import dedent

from maccarone.preprocessor import (
    PresentPiece,
    MissingPiece,
    raw_source_to_pieces,
    raw_pieces_to_tagged_input,
    tagged_output_to_completed_pieces,
)

@pytest.mark.parametrize("input, expected", [
    (
        "",
        [PresentPiece("")],
    ),
    (
        "This is a normal string",
        [PresentPiece("This is a normal string")],
    ),
    (
        "This string has #<<one>> missing piece",
        [
            PresentPiece("This string has "),
            MissingPiece("", "one"),
            PresentPiece(" missing piece"),
        ],
    ),
    (
        "This string has #<<one>> and #<<two>> missing pieces",
        [
            PresentPiece("This string has "),
            MissingPiece("", "one"),
            PresentPiece(" and "),
            MissingPiece("", "two"),
            PresentPiece(" missing pieces"),
        ],
    ),
])
def test_raw_source_to_pieces(input, expected):
    assert list(raw_source_to_pieces(input)) == expected

@pytest.mark.parametrize("raw_pieces, expected", [
    (
        [
            PresentPiece("\ndef add_two_numbers(x, y):\n    "),
            MissingPiece("    ", "add the args"),
            PresentPiece("\n\n"),
            MissingPiece("", "add two numbers from command line args, using argparse"),
            PresentPiece("\n"),
        ],
        dedent("""
        def add_two_numbers(x, y):
            # <write_this id="0">
            # add the args
            # </>
        
        # <write_this id="1">
        # add two numbers from command line args, using argparse
        # </>
        """),
    ),
])
def test_raw_source_to_tagged_input(raw_pieces, expected):
    assert raw_pieces_to_tagged_input(raw_pieces) == expected

@pytest.mark.parametrize("tagged, expected", [
    (
        '<completed id="0">\ndef add_two_numbers(x, y):\n    return x + y\n</>\n',
        {0: 'def add_two_numbers(x, y):\n    return x + y\n'}
    ),
    (
        '<completed id="1">\ndef subtract_two_numbers(x, y):\n    return x - y\n</>\n',
        {1: 'def subtract_two_numbers(x, y):\n    return x - y\n'}
    ),
    (
        '<completed id="1">\nfoo\n</>\n<completed id="2">\ndef multiply_two_numbers(x, y):\n    return x * y\n</>\n',
        {
            1: "foo\n",
            2: 'def multiply_two_numbers(x, y):\n    return x * y\n'
        }
    ),
])
def test_tagged_output_to_completed_pieces(tagged, expected):
    assert tagged_output_to_completed_pieces(tagged) == expected



from tempfile import TemporaryDirectory

from maccarone.caching import (
    DiskCache,
    CacheKeyMissingError,
)

def test_disk_cache_set_get():
    with TemporaryDirectory() as cache_dir:
        cache = DiskCache(cache_dir)

        key = {"a": 1, "b": 2}
        value = {"c": 3, "d": 4}

        try:
            cache.get(key)
        except CacheKeyMissingError:
            pass
        else:
            assert False, "expected CacheKeyMissingError"

        cache.set(key, value)

        assert cache.get(key) == value



import glob
import logging

from argparse import (
    ArgumentParser,
    Namespace,
)

from maccarone.preprocessor import preprocess_maccarone

logger = logging.getLogger(__name__)

def preprocess(mn_path: str, suffix=".mn.py") -> None:
    # replace suffix with `.py`
    assert mn_path.endswith(suffix)

    py_path = mn_path[:-len(suffix)] + ".py"

    # produce Python source
    logger.info("preprocessing %s → %s", mn_path, py_path)

    with open(mn_path, "rt") as mn_file:
        mn_source = mn_file.read()
        py_source = preprocess_maccarone(mn_source)

    with open(py_path, "wt") as file:
        file.write(py_source)

def main(root_path: str):
    """Preprocess maccarone files into Python."""

    for path in glob.glob(root_path + "/**/*.mn.py", recursive=True):
        preprocess(path)

def parse_args() -> Namespace:
    #<<use argparse to handle `script ROOT_PATH`>>

def script_main():
    logging.basicConfig(level=logging.INFO)

    return main(**vars(parse_args()))

if __name__ == "__main__":
    script_main()




from logging import getLogger

logger = getLogger(__name__)

try:
    import stale_maccarone.enable

    stale_maccarone.enable.enable_py_string_matching = False
    stale_maccarone.enable.fullname_pattern = r"maccarone\..*"
except ImportError:
    logger.debug("no stale_maccarone; assuming source is already preprocessed")

from maccarone.loader import enable
